@@ -52,9 +52,6 @@
 //C- | TO ANY WARRANTY OF NON-INFRINGEMENT, OR ANY IMPLIED WARRANTY OF
 //C- | MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.
 //C- +------------------------------------------------------------------
-// 
-// $Id: GUnicode.cpp,v 1.12 2007/03/25 20:48:32 leonb Exp $
-// $Name: release_3_5_20_2 $
 
 #ifdef HAVE_CONFIG_H
 # include "config.h"
@@ -64,11 +61,13 @@
 #endif
 
 #include "GString.h"
+
+#include <stddef.h>
+
 #if HAS_ICONV
 #include <iconv.h>
 #endif
 
-
 #ifdef HAVE_NAMESPACES
 namespace DJVU {
 # ifdef NOT_DEFINED // Just to fool emacs c++ mode
@@ -188,14 +187,14 @@ protected:
   virtual void set_remainder( const GP<Unicode> &remainder );
   virtual GP<Unicode> get_remainder(void) const;
 };
-// static unsigned long UTF8toUCS4(unsigned char const *&,void const * const);
-static unsigned long xUTF16toUCS4(unsigned short const *&s,void const * const);
-static unsigned long UTF16BEtoUCS4(unsigned char const *&s,void const * const);
-static unsigned long UTF16LEtoUCS4(unsigned char const *&s,void const * const);
-static unsigned long UCS4BEtoUCS4(unsigned char const *&s,void const * const);
-static unsigned long UCS4LEtoUCS4(unsigned char const *&s,void const * const);
-static unsigned long UCS4_3412toUCS4(unsigned char const *&s,void const * const);
-static unsigned long UCS4_2143toUCS4(unsigned char const *&s,void const * const);
+// static uint32_t UTF8toUCS4(unsigned char const *&,void const * const);
+static uint32_t xUTF16toUCS4(uint16_t const *&s,void const * const);
+static uint32_t UTF16BEtoUCS4(unsigned char const *&s,void const * const);
+static uint32_t UTF16LEtoUCS4(unsigned char const *&s,void const * const);
+static uint32_t UCS4BEtoUCS4(unsigned char const *&s,void const * const);
+static uint32_t UCS4LEtoUCS4(unsigned char const *&s,void const * const);
+static uint32_t UCS4_3412toUCS4(unsigned char const *&s,void const * const);
+static uint32_t UCS4_2143toUCS4(unsigned char const *&s,void const * const);
 
 GP<GStringRep>
 GStringRep::Unicode::create(const unsigned int sz)
@@ -352,12 +351,12 @@ GStringRep::Unicode::create(
             size_t pleft=6*ptrleft+1;
             GPBuffer<char> gutf8buf(utf8buf,pleft);
             char *p=utf8buf;
-            unsigned char const *last=ptr;
-            for(;iconv_adaptor(iconv, cv, (char**)&ptr, &ptrleft, &p, &pleft);last=ptr) 
-              EMPTY_LOOP;
+            char *nptr = (char*)ptr;
+            while(iconv_adaptor(iconv, cv, &nptr, &ptrleft, &p, &pleft)) 
+              ptr = (unsigned char*)nptr;
             iconv_close(cv);
-            retval=create(utf8buf,(size_t)last-(size_t)buf,t);
-            retval->set_remainder(last,(size_t)eptr-(size_t)last,e);
+            retval=create(utf8buf,(size_t)ptr-(size_t)buf,t);
+            retval->set_remainder(ptr,(size_t)eptr-(size_t)ptr,e);
           }
         }
       }else
@@ -395,9 +394,9 @@ GStringRep::Unicode::create(
       case XUCS4_2143:
       case XUCS4_3412:
       {
-        for(unsigned long w;
-          (eptr<xeptr)&&(w=*(unsigned long const *)eptr);
-          eptr+=sizeof(unsigned long))
+        for(uint32_t w;
+          (eptr<xeptr)&&(w=*(uint32_t const *)eptr);
+          eptr+=sizeof(uint32_t))
         {
           maxutf8size+=(w>0x7f)?6:1;
         }
@@ -407,9 +406,9 @@ GStringRep::Unicode::create(
       case XUTF16BE:
       case XUTF16LE:
       {
-        for(unsigned short w;
-          (eptr<xeptr)&&(w=*(unsigned short const *)eptr);
-          eptr+=sizeof(unsigned short))
+        for(uint16_t w;
+          (eptr<xeptr)&&(w=*(uint16_t const *)eptr);
+          eptr+=sizeof(uint16_t))
         {
           maxutf8size+=3;
         }
@@ -436,13 +435,14 @@ GStringRep::Unicode::create(
       unsigned char *optr=utf8buf;
       int len=0;
       unsigned char const *iptr=(unsigned char *)buf;
-      unsigned long w;
+      uint16_t const *sptr=(uint16_t *)buf;
+      uint32_t w;
       switch(t)
       {
         case XUCS4:
           for(;
-            (iptr<eptr)&&(w=*(unsigned long const *)iptr);
-            len++,iptr+=sizeof(unsigned long const))
+            (iptr<eptr)&&(w=*(uint32_t const *)iptr);
+            len++,iptr+=sizeof(uint32_t const))
           {
             optr=UCS4toUTF8(w,optr);
           }
@@ -472,9 +472,7 @@ GStringRep::Unicode::create(
           }
           break;
         case XUTF16:
-          for(;
-            (w=xUTF16toUCS4((unsigned short const*&)iptr,eptr));
-            len++)
+          for(;(w=xUTF16toUCS4(sptr,eptr));len++)
           {
             optr=UCS4toUTF8(w,optr);
           }
@@ -532,14 +530,14 @@ GStringRep::Unicode::create(
   return gretval;
 }
 
-static unsigned long
-xUTF16toUCS4(unsigned short const *&s,void const * const eptr)
+static uint32_t
+xUTF16toUCS4(uint16_t const *&s,void const * const eptr)
 {
-  unsigned long U=0;
-  unsigned short const * const r=s+1;
+  uint32_t U=0;
+  uint16_t const * const r=s+1;
   if(r <= eptr)
   {
-    unsigned long const W1=s[0];
+    uint32_t const W1=s[0];
     if((W1<0xD800)||(W1>0xDFFF))
     {
       if((U=W1))
@@ -548,10 +546,10 @@ xUTF16toUCS4(unsigned short const *&s,void const * const eptr)
       }
     }else if(W1<=0xDBFF)
     {
-      unsigned short const * const rr=r+1;
+      uint16_t const * const rr=r+1;
       if(rr <= eptr)
       {
-        unsigned long const W2=s[1];
+        uint32_t const W2=s[1];
         if(((W2>=0xDC00)||(W2<=0xDFFF))&&((U=(0x1000+((W1&0x3ff)<<10))|(W2&0x3ff))))
         {
           s=rr;
@@ -566,17 +564,17 @@ xUTF16toUCS4(unsigned short const *&s,void const * const eptr)
   return U;
 }
 
-static unsigned long
+static uint32_t
 UTF16BEtoUCS4(unsigned char const *&s,void const * const eptr)
 {
-  unsigned long U=0;
+  uint32_t U=0;
   unsigned char const * const r=s+2;
   if(r <= eptr)
   {
-    unsigned long const C1MSB=s[0];
+    uint32_t const C1MSB=s[0];
     if((C1MSB<0xD8)||(C1MSB>0xDF))
     {
-      if((U=((C1MSB<<8)|((unsigned long)s[1]))))
+      if((U=((C1MSB<<8)|((uint32_t)s[1]))))
       {
         s=r;
       }
@@ -585,15 +583,15 @@ UTF16BEtoUCS4(unsigned char const *&s,void const * const eptr)
       unsigned char const * const rr=r+2;
       if(rr <= eptr)
       {
-        unsigned long const C2MSB=s[2];
+        uint32_t const C2MSB=s[2];
         if((C2MSB>=0xDC)||(C2MSB<=0xDF))
         {
-          U=0x10000+((unsigned long)s[1]<<10)+(unsigned long)s[3]
+          U=0x10000+((uint32_t)s[1]<<10)+(uint32_t)s[3]
             +(((C1MSB<<18)|(C2MSB<<8))&0xc0300);
           s=rr;
         }else
         {
-          U=(unsigned int)(-1)-((C1MSB<<8)|((unsigned long)s[1]));
+          U=(unsigned int)(-1)-((C1MSB<<8)|((uint32_t)s[1]));
           s=r;
         }
       }
@@ -602,17 +600,17 @@ UTF16BEtoUCS4(unsigned char const *&s,void const * const eptr)
   return U;
 }
 
-static unsigned long
+static uint32_t
 UTF16LEtoUCS4(unsigned char const *&s,void const * const eptr)
 {
-  unsigned long U=0;
+  uint32_t U=0;
   unsigned char const * const r=s+2;
   if(r <= eptr)
   {
-    unsigned long const C1MSB=s[1];
+    uint32_t const C1MSB=s[1];
     if((C1MSB<0xD8)||(C1MSB>0xDF))
     {
-      if((U=((C1MSB<<8)|((unsigned long)s[0]))))
+      if((U=((C1MSB<<8)|((uint32_t)s[0]))))
       {
         s=r;
       }
@@ -621,15 +619,15 @@ UTF16LEtoUCS4(unsigned char const *&s,void const * const eptr)
       unsigned char const * const rr=r+2;
       if(rr <= eptr)
       {
-        unsigned long const C2MSB=s[3];
+        uint32_t const C2MSB=s[3];
         if((C2MSB>=0xDC)||(C2MSB<=0xDF))
         {
-          U=0x10000+((unsigned long)s[0]<<10)+(unsigned long)s[2]
+          U=0x10000+((uint32_t)s[0]<<10)+(uint32_t)s[2]
             +(((C1MSB<<18)|(C2MSB<<8))&0xc0300);
           s=rr;
         }else
         {
-          U=(unsigned int)(-1)-((C1MSB<<8)|((unsigned long)s[1]));
+          U=(unsigned int)(-1)-((C1MSB<<8)|((uint32_t)s[1]));
           s=r;
         }
       }
@@ -638,14 +636,14 @@ UTF16LEtoUCS4(unsigned char const *&s,void const * const eptr)
   return U;
 }
 
-static unsigned long
+static uint32_t
 UCS4BEtoUCS4(unsigned char const *&s,void const * const eptr)
 {
-  unsigned long U=0;
+  uint32_t U=0;
   unsigned char const * const r=s+4;
   if(r<=eptr)
   {
-    U=(((((((unsigned long)s[0]<<8)|(unsigned long)s[1])<<8)|(unsigned long)s[2])<<8)|(unsigned long)s[3]);
+    U=(((((((uint32_t)s[0]<<8)|(uint32_t)s[1])<<8)|(uint32_t)s[2])<<8)|(uint32_t)s[3]);
     if(U)
     {
       s=r;
@@ -654,14 +652,14 @@ UCS4BEtoUCS4(unsigned char const *&s,void const * const eptr)
   return U;
 }
 
-static unsigned long
+static uint32_t
 UCS4LEtoUCS4(unsigned char const *&s,void const * const eptr)
 {
-  unsigned long U=0;
+  uint32_t U=0;
   unsigned char const * const r=s+4;
   if(r<=eptr)
   {
-    U=(((((((unsigned long)s[3]<<8)|(unsigned long)s[2])<<8)|(unsigned long)s[1])<<8)|(unsigned long)s[0]);
+    U=(((((((uint32_t)s[3]<<8)|(uint32_t)s[2])<<8)|(uint32_t)s[1])<<8)|(uint32_t)s[0]);
     if(U)
     {
       s=r;
@@ -670,14 +668,14 @@ UCS4LEtoUCS4(unsigned char const *&s,void const * const eptr)
   return U;
 }
 
-static unsigned long
+static uint32_t
 UCS4_2143toUCS4(unsigned char const *&s,void const * const eptr)
 {
-  unsigned long U=0;
+  uint32_t U=0;
   unsigned char const * const r=s+4;
   if(r<=eptr)
   {
-    U=(((((((unsigned long)s[1]<<8)|(unsigned long)s[0])<<8)|(unsigned long)s[3])<<8)|(unsigned long)s[2]);
+    U=(((((((uint32_t)s[1]<<8)|(uint32_t)s[0])<<8)|(uint32_t)s[3])<<8)|(uint32_t)s[2]);
     if(U)
     {
       s=r;
@@ -686,14 +684,14 @@ UCS4_2143toUCS4(unsigned char const *&s,void const * const eptr)
   return U;
 }
 
-static unsigned long
+static uint32_t
 UCS4_3412toUCS4(unsigned char const *&s,void const * const eptr)
 {
-  unsigned long U=0;
+  uint32_t U=0;
   unsigned char const * const r=s+4;
   if(r<=eptr)
   {
-    U=(((((((unsigned long)s[2]<<8)|(unsigned long)s[3])<<8)|(unsigned long)s[0])<<8)|(unsigned long)s[1]);
+    U=(((((((uint32_t)s[2]<<8)|(uint32_t)s[3])<<8)|(uint32_t)s[0])<<8)|(uint32_t)s[1]);
     if(U)
     {
       s=r;
