@@ -15,7 +15,6 @@
 // GNU General Public License for more details.
 // -------------------------------------------------------------------
 */
-/* $Id: miniexp.cpp,v 1.16 2007/12/02 03:58:29 leonb Exp $ */
 
 #ifdef HAVE_CONFIG_H
 # include "config.h"
@@ -24,6 +23,7 @@
 # pragma implementation "miniexp.h"
 #endif
 
+#include <stddef.h>
 #include <stdlib.h>
 #include <stdio.h>
 #include <ctype.h>
@@ -259,8 +259,9 @@ collect_free(block_t *b, void **&freelist, int &count, bool destroy)
       for (unsigned int i=1; i<nptrs_chunk/2; i++)
         if (! c[i])
           {
-            if (destroy && m[i+i]==m[i+i+1]) 
-              delete (miniobj_t*)m[i+i];
+            miniobj_t *obj = (miniobj_t*)m[i+i];
+            if (destroy && obj && m[i+i]==m[i+i+1]) 
+              obj->destroy();
             m[i+i] = (void*)freelist;
             m[i+i+1] = 0;
             freelist = &m[i+i];
@@ -312,42 +313,109 @@ new_obj_block(void)
 }
 
 #if defined(__GNUC__) && (__GNUC__ >= 3)
-static void gc_mark_object(void **v)
-  __attribute__((noinline));
+static void gc_mark_object(void **v) __attribute__((noinline));
 #else
 static void gc_mark_object(void **v);
 #endif
 
+static bool
+gc_mark_check(void *p)
+{
+  if (((size_t)p) & 2)
+    return false;
+  void **v = (void**)(((size_t)p) & ~(size_t)3); 
+  if (! v)
+    return false;
+  char *m = markbyte(v);
+  if (*m)
+    return false;
+  *m = 1;
+  if (! (((size_t)p) & 1))
+    return true;
+  gc_mark_object((void**)v);
+  return false;
+}
+
+static void
+gc_mark_pair(void **v)
+{
+#ifndef MINIEXP_POINTER_REVERSAL
+  // This is a simple recursive code.
+  // Despite the tail recursion for the cdrs,
+  // it consume a stack space that grows like
+  // the longest chain of cars.
+  for(;;)
+    {
+      if (gc_mark_check(v[0]))
+        gc_mark_pair((void**)v[0]);
+      if (! gc_mark_check(v[1]))
+        break;
+      v = (void**)v[1];
+    }
+#else
+  // This is the classic pointer reversion code
+  // It saves stack memory by temporarily reversing the pointers. 
+  // This is a bit slower because of all these nonlocal writes.
+  // But it could be useful for memory-starved applications.
+  // That makes no sense for most uses of miniexp.
+  // I leave the code here because of its academic interest.
+  void **w = 0;
+ docar:
+  if (gc_mark_check(v[0]))
+    { // reverse car pointer
+      void **p = (void**)v[0];
+      v[0] = (void*)w;
+      w = (void**)(((size_t)v)|(size_t)1);
+      v = p;
+      goto docar;
+    }
+ docdr:
+  if (gc_mark_check(v[1]))
+    { // reverse cdr pointer
+      void **p = (void**)v[1];
+      v[1] = (void*)w;
+      w = v;
+      v = p;
+      goto docar;
+    }
+ doup:
+  if (w)
+    {
+      if (((size_t)w)&1)
+        { // undo car reversion
+          void **p = (void**)(((size_t)w)&~(size_t)1);
+          w = (void**)p[0];
+          p[0] = (void*)v;
+          v = p;
+          goto docdr;
+        }
+      else
+        { // undo cdr reversion
+          void **p = w;
+          w = (void**)p[1];
+          p[1] = (void*)v;
+          v = p;
+          goto doup;
+        }
+    }
+#endif
+}
+
 static void
 gc_mark(miniexp_t *pp)
 {
-  for(;;)
-    {
-      miniexp_t p = *pp;
-      if (((size_t)p) & 2) return;
-      void **v = (void**)(((size_t)p) & ~(size_t)3);
-      if (! v) return;
-      char *m = markbyte(v);
-      if (*m) return;
-      (*m) = 1;
-      if (((size_t)p) & 1)
-        { // object
-          gc_mark_object(v);
-          return;
-        }
-      else
-        { // pair
-          gc_mark((miniexp_t*)&v[0]);
-          pp = (miniexp_t*)&v[1];
-        }
-    }
+  void **v = (void**)*pp;
+  if (gc_mark_check((void**)*pp))
+    gc_mark_pair(v);
 }
 
 static void
 gc_mark_object(void **v)
 {
+  ASSERT(v[0] == v[1]);
   miniobj_t *obj = (miniobj_t*)v[0];
-  if (obj) obj->mark(gc_mark);
+  if (obj) 
+    obj->mark(gc_mark);
 }
 
 static void
@@ -366,9 +434,9 @@ gc_run(void)
       // mark
       minivar_t::mark(gc_mark);
       for (int i=0; i<recentsize; i++)
-        {
-          miniexp_t p = (miniexp_t)gc.recent[i];
-          gc_mark(&p);
+        { // extra cast for strict aliasing rules?
+          char *s = (char*)&gc.recent[i];
+          gc_mark((miniexp_t*)s);
         }
       // sweep
       gc.objs_free = gc.pairs_free = 0;
@@ -395,13 +463,12 @@ gc_alloc_pair(void *a, void *d)
         new_pair_block();
     }
   else if (gc.debug)
-    minilisp_gc();
+    gc_run();
   void **p = gc.pairs_freelist;
   gc.pairs_freelist = (void**)p[0];
   gc.pairs_free -= 1;
   p[0] = a;
   p[1] = d;
-  gc.recent[(++gc.recentindex) & (recentsize-1)] = p;
   return p;
 }
 
@@ -415,12 +482,11 @@ gc_alloc_object(void *obj)
         new_obj_block();
     }
   else if (gc.debug)
-    minilisp_gc();
+    gc_run();
   void **p = gc.objs_freelist;
   gc.objs_freelist = (void**)p[0];
   gc.objs_free -= 1;
   p[0] = p[1] = obj;
-  gc.recent[(++gc.recentindex) & (recentsize-1)] = p;
   return p;
 }
 
@@ -621,9 +687,8 @@ miniexp_nth(int n, miniexp_t l)
 miniexp_t 
 miniexp_cons(miniexp_t a, miniexp_t d)
 {
-  gc.recent[(gc.recentindex+1) & (recentsize-1)] = (void**)a;
-  gc.recent[(gc.recentindex+2) & (recentsize-1)] = (void**)d;
   miniexp_t r = (miniexp_t)gc_alloc_pair((void*)a, (void*)d); 
+  gc.recent[(++gc.recentindex) & (recentsize-1)] = (void**)r;
   return r;
 }
 
@@ -685,12 +750,18 @@ miniobj_t::mark(minilisp_mark_t*)
 {
 }
 
+void 
+miniobj_t::destroy()
+{
+  delete this;
+}
+
 char *
 miniobj_t::pname() const
 {
   const char *cname = miniexp_to_name(classof());
   char *res = new char[strlen(cname)+24];
-  sprintf(res,"#<%s:%p>",cname,this);
+  sprintf(res,"#%s:<%p>",cname,this);
   return res;
 }
 
@@ -698,7 +769,9 @@ miniexp_t
 miniexp_object(miniobj_t *obj)
 {
   void **v = gc_alloc_object((void*)obj);
-  return (miniexp_t)(((size_t)v)|(size_t)1);
+  v = (void**)(((size_t)v)|((size_t)1));
+  gc.recent[(++gc.recentindex) & (recentsize-1)] = v;
+  return (miniexp_t)(v);
 }
 
 miniexp_t 
@@ -821,10 +894,9 @@ print_c_string(const char *s, char *d, bool eightbits)
 char *
 ministring_t::pname() const
 {
-  bool eightbits = !minilisp_print_7bits;
-  int n = print_c_string(s, 0, eightbits);
+  int n = print_c_string(s, 0, true);
   char *d = new char[n];
-  if (d) print_c_string(s, d, eightbits);
+  if (d) print_c_string(s, d, true);
   return d;
 }
 
@@ -889,61 +961,132 @@ miniexp_concat(miniexp_t p)
 /* INPUT/OUTPUT                                       */
 /* -------------------------------------------------- */
 
-extern "C" { 
-  // SunCC needs this to be defined inside extern "C" { ... }
-  // Beware the difference between extern "C" {...} and extern "C".
-  miniexp_t (*minilisp_macrochar_parser[128])(void); 
+BEGIN_ANONYMOUS_NAMESPACE
+
+struct CompatCounter 
+{
+  static int count;
+  CompatCounter() { count += 1; }
+  ~CompatCounter() { count -= 1; }
+};
+
+int CompatCounter::count = 0;
+
+END_ANONYMOUS_NAMESPACE
+
+static int
+compat_puts(const char *s)
+{
+  CompatCounter count;
+  return miniexp_io.fputs(&miniexp_io, s);
 }
 
-/* --------- OUTPUT */
+static int
+compat_getc()
+{
+  CompatCounter count;
+  return miniexp_io.fgetc(&miniexp_io);
+}
 
-static FILE *outputfile;
+static int
+compat_ungetc(int c)
+{
+  CompatCounter count;
+  return miniexp_io.ungetc(&miniexp_io, c);
+}
 
 static int 
-stdio_puts(const char *s)
+stdio_fputs(miniexp_io_t *io, const char *s)
 {
-  if (!outputfile)
-    outputfile = stdout;
-  return fputs(s, outputfile);
-  return EOF;
+  if (io == &miniexp_io && !CompatCounter::count)
+    return (*minilisp_puts)(s); /* compatibility hack */
+  FILE *f = (io->data[1]) ? (FILE*)(io->data[1]) : stdout;
+  return ::fputs(s, f);
 }
 
-int (*minilisp_puts)(const char *s) = stdio_puts;
+static int 
+stdio_fgetc(miniexp_io_t *io)
+{
+  if (io == &miniexp_io && !CompatCounter::count)
+    return (*minilisp_getc)(); /* compatibility hack */
+  FILE *f = (io->data[0]) ? (FILE*)(io->data[0]) : stdin;
+  return ::getc(f);
+}
 
-int minilisp_print_7bits = 1;
+static int 
+stdio_ungetc(miniexp_io_t *io, int c)
+{
+  if (io == &miniexp_io && !CompatCounter::count)
+    return (*minilisp_ungetc)(c); /* compatibility hack */
+  FILE *f = (io->data[0]) ? (FILE*)(io->data[0]) : stdin;
+  return ::ungetc(c, f);
+}
+
+extern "C" 
+{ 
+  // SunCC needs this to be defined inside extern "C" { ... }
+  // Beware the difference between extern "C" {...} and extern "C".
+  miniexp_t (*minilisp_macrochar_parser[128])(void);
+  minivar_t minilisp_macroqueue;
+  int minilisp_print_7bits;
+}
+
+
+miniexp_io_t miniexp_io = { 
+  stdio_fputs, stdio_fgetc, stdio_ungetc, { 0, 0, 0, 0 },
+  (int*)&minilisp_print_7bits,
+  (miniexp_macrochar_t*)minilisp_macrochar_parser, 
+  (minivar_t*)&minilisp_macroqueue,
+  0
+};  
+
+int (*minilisp_puts)(const char *) = compat_puts;
+int (*minilisp_getc)(void) = compat_getc;
+int (*minilisp_ungetc)(int) = compat_ungetc;
 
 void 
-minilisp_set_output(FILE *f)
+miniexp_io_init(miniexp_io_t *io)
 {
-  outputfile = f;
-  minilisp_puts = stdio_puts;
+  io->fputs = stdio_fputs;
+  io->fgetc = stdio_fgetc;
+  io->ungetc = stdio_ungetc;
+  io->data[0] = io->data[1] = io->data[2] = io->data[3] = 0;
+  io->p_print7bits = (int*)&minilisp_print_7bits;;
+  io->p_macrochar = (miniexp_macrochar_t*)minilisp_macrochar_parser;
+  io->p_macroqueue = (minivar_t*)&minilisp_macroqueue;
+  io->p_reserved = 0;
 }
 
-static bool
-must_quote_symbol(const char *s)
+void 
+miniexp_io_set_output(miniexp_io_t* io, FILE *f)
 {
-  int c;
-  const char *r = s;
-  while ((c = *r++))
-    if (c=='(' || c==')' || c=='\"' || c=='|' || 
-        isspace(c) || !isascii(c) || !isprint(c) ||
-        minilisp_macrochar_parser[c] )
-      return true;
-  char *end;
-  strtol(s,&end,0);
-  return !(*end);
+  io->fputs = stdio_fputs;
+  io->data[1] = f;
 }
 
+void 
+miniexp_io_set_input(miniexp_io_t* io, FILE *f)
+{
+  io->fgetc = stdio_fgetc;
+  io->ungetc = stdio_ungetc;
+  io->data[0] = f;
+}
+
+
+/* ---- OUTPUT */
+
 BEGIN_ANONYMOUS_NAMESPACE
 
 struct printer_t 
 {
   int tab;
   bool dryrun;
-  printer_t() : tab(0), dryrun(false) {}
+  miniexp_io_t *io;
+  printer_t(miniexp_io_t *io) : tab(0), dryrun(false), io(io) {}
   void mlput(const char *s);
   void mltab(int n);
   void print(miniexp_t p);
+  bool must_quote_symbol(const char *s);
   virtual miniexp_t begin() { return miniexp_nil; }
   virtual bool newline() { return false; }
   virtual void end(miniexp_t) { }
@@ -954,7 +1097,7 @@ void
 printer_t::mlput(const char *s)
 {
   if (! dryrun)
-    minilisp_puts(s);
+    io->fputs(io, s);
   while (*s)
     if (*s++ == '\n')
       tab = 0;
@@ -971,6 +1114,24 @@ printer_t::mltab(int n)
     mlput(" ");
 }
 
+bool
+printer_t::must_quote_symbol(const char *s)
+{
+  int c;
+  const char *r = s;
+  while ((c = *r++))
+    if (c=='(' || c==')' || c=='\"' || c=='|' || 
+        isspace(c) || !isascii(c) || !isprint(c) ||
+        (c >= 0 && c < 128 && io->p_macrochar && io->p_macrochar[c]) )
+      return true;
+  char *end;
+#ifdef __GNUC__
+  long junk __attribute__ ((unused)) =
+#endif
+  strtol(s, &end, 0);
+  return (!*end);
+}
+
 void
 printer_t::print(miniexp_t p)
 {
@@ -982,7 +1143,7 @@ printer_t::print(miniexp_t p)
     }
   else if (p == miniexp_dummy)
     {
-      mlput("#<dummy>");
+      mlput("#dummy");
     }
   else if (miniexp_numberp(p))
     {
@@ -997,6 +1158,16 @@ printer_t::print(miniexp_t p)
       mlput(s);
       if (needquote) mlput("|");
     }
+  else if (miniexp_stringp(p))
+    {
+      const char *s = miniexp_to_str(p);
+      bool print7bits = (io->p_print7bits && *io->p_print7bits);
+      int n = print_c_string(s, 0, !print7bits);
+      char *d = new char[n];
+      if (d) print_c_string(s, d, !print7bits);
+      mlput(d);
+      delete [] d;
+    }
   else if (miniexp_objectp(p))
     {
       miniobj_t *obj = miniexp_to_obj(p);
@@ -1021,7 +1192,7 @@ printer_t::print(miniexp_t p)
       while (miniexp_consp(p))
         {
           skip -= 1;
-          if (multiline || newline() && skip<0 && tab>indent)
+	  if (multiline || (newline() && skip<0 && tab>indent))
             {
               mlput("\n"); 
               mltab(indent); 
@@ -1041,7 +1212,7 @@ printer_t::print(miniexp_t p)
       if (p)
         {
           skip -= 1;
-          if (multiline || newline() && skip<0 && tab>indent)
+	  if (multiline || (newline() && skip<0 && tab>indent))
             {
               mlput("\n"); 
               mltab(indent); 
@@ -1062,6 +1233,7 @@ struct pprinter_t : public printer_t
 {
   int width;
   minivar_t l;
+  pprinter_t(miniexp_io_t *io) : printer_t(io) {}
   virtual miniexp_t begin();
   virtual bool newline();
   virtual void end(miniexp_t);
@@ -1114,28 +1286,28 @@ pprinter_t::end(miniexp_t p)
 END_ANONYMOUS_NAMESPACE
 
 miniexp_t 
-miniexp_prin(miniexp_t p)
+miniexp_prin_r(miniexp_io_t *io, miniexp_t p)
 {
   minivar_t xp = p;
-  printer_t printer;
+  printer_t printer(io);
   printer.print(p);
   return p;
 }
 
 miniexp_t 
-miniexp_print(miniexp_t p)
+miniexp_print_r(miniexp_io_t *io, miniexp_t p)
 {
   minivar_t xp = p;
-  miniexp_prin(p);
-  minilisp_puts("\n");
+  miniexp_prin_r(io, p);
+  io->fputs(io, "\n");
   return p;
 }
 
 miniexp_t 
-miniexp_pprin(miniexp_t p, int width)
+miniexp_pprin_r(miniexp_io_t *io, miniexp_t p, int width)
 {  
   minivar_t xp = p;
-  pprinter_t printer;
+  pprinter_t printer(io);
   printer.width = width;
   // step1 - measure lengths into list <l>
   printer.tab = 0;
@@ -1152,34 +1324,36 @@ miniexp_pprin(miniexp_t p, int width)
 }
 
 miniexp_t 
-miniexp_pprint(miniexp_t p, int width)
+miniexp_pprint_r(miniexp_io_t *io, miniexp_t p, int width)
 {
-  miniexp_pprin(p, width);
-  minilisp_puts("\n");
+  miniexp_pprin_r(io, p, width);
+  io->fputs(io, "\n");
   return p;
 }
 
-/* --------- PNAME */
 
-static struct { 
-  char *b; int l; int m; 
-} pname_data;
+/* ---- PNAME */
 
 static int
-pname_puts(const char *s)
+pname_fputs(miniexp_io_t *io, const char *s)
 {
-  int x = strlen(s);
-  if (pname_data.l + x >= pname_data.m)
+  char *b = (char*)(io->data[0]);
+  size_t l = (size_t)(io->data[2]);
+  size_t m = (size_t)(io->data[3]);
+  size_t x = strlen(s);
+  if (l + x >= m)
     {
-      int nm = pname_data.l + x + 256;
+      size_t nm = l + x + 256;
       char *nb = new char[nm+1];
-      memcpy(nb, pname_data.b, pname_data.l);
-      delete [] pname_data.b;
-      pname_data.m = nm;
-      pname_data.b = nb;
+      memcpy(nb, b, l);
+      delete [] b;
+      b = nb;
+      m = nm;
     }
-  strcpy(pname_data.b + pname_data.l, s);
-  pname_data.l += x;
+  strcpy(b + l, s);
+  io->data[0] = (void*)(b);
+  io->data[2] = (void*)(l + x);
+  io->data[3] = (void*)(m);
   return x;
 }
 
@@ -1187,71 +1361,29 @@ miniexp_t
 miniexp_pname(miniexp_t p, int width)
 {
   minivar_t r;
-  int (*saved)(const char*) = minilisp_puts;
-  pname_data.b = 0;
-  pname_data.m = pname_data.l = 0;
+  miniexp_io_t io;
+  miniexp_io_init(&io);
+  io.fputs = pname_fputs;
+  io.data[0] = io.data[2] = io.data[3] = 0;
   try
     {
-      minilisp_puts = pname_puts;
       if (width > 0)
-        miniexp_pprin(p, width);
+        miniexp_pprin_r(&io, p, width);
       else
-        miniexp_prin(p);
-      minilisp_puts = saved;
-      r = miniexp_string(pname_data.b);
-      delete [] pname_data.b;
-      pname_data.b = 0;
+        miniexp_prin_r(&io, p);
+      if (io.data[0])
+        r = miniexp_string((const char*)io.data[0]);
+      delete [] (char*)(io.data[0]);
     }
   catch(...)
     {
-      minilisp_puts = saved;
-      delete [] pname_data.b;
-      pname_data.b = 0;
+      delete [] (char*)(io.data[0]);
     }
   return r;
 }
 
 
-
-/* --------- INPUT */
-
-static FILE *inputfile;
-static minivar_t inputqueue;
-
-static int
-stdio_getc(void)
-{
-  if (!inputfile)
-    inputfile = stdin;
-  return getc(inputfile);
-}
-
-static int
-stdio_ungetc(int c)
-{
-  if (inputfile && c>=0)
-    return ungetc(c, inputfile);
-  return EOF;
-}
-
-int (*minilisp_getc)(void) = stdio_getc;
-
-int (*minilisp_ungetc)(int c) = stdio_ungetc;
-
-void 
-minilisp_set_input(FILE *f)
-{
-  inputfile = f;
-  minilisp_getc = stdio_getc;
-  minilisp_ungetc = stdio_ungetc;
-}
-
-static void
-skip_blank(int &c)
-{
-  while (isspace(c))
-    c = minilisp_getc();
-}
+/* ---- INPUT */
 
 static void
 append(int c, char* &s, int &l, int &m)
@@ -1269,49 +1401,56 @@ append(int c, char* &s, int &l, int &m)
   s[l] = 0;
 }
 
+static void
+skip_blank(miniexp_io_t *io, int &c)
+{
+  while (isspace(c))
+    c = io->fgetc(io);
+}
+
 static miniexp_t
-read_error(int &c)
+read_error(miniexp_io_t *io, int &c)
 {
   while (c!=EOF && c!='\n')
-    c = minilisp_getc();
+    c = io->fgetc(io);
   return miniexp_dummy;
 }
 
 static miniexp_t
-read_c_string(int &c)
+read_c_string(miniexp_io_t *io, int &c)
 {
   miniexp_t r;
   char *s = 0;
   int l = 0;
   int m = 0;
   ASSERT(c == '\"');
-  c = minilisp_getc();
+  c = io->fgetc(io);
   for(;;)
     {
-      if (c==EOF || isascii(c) && !isprint(c))
-        return read_error(c);
+      if (c==EOF || (isascii(c) && !isprint(c)))
+        return read_error(io, c);
       else if (c=='\"')
         break;
       else if (c=='\\')
         {
-          c = minilisp_getc();
+          c = io->fgetc(io);
           if (c == '\n')
             {
-              c = minilisp_getc();
+              c = io->fgetc(io);
               continue;
             }
           else if (c>='0' && c<='7')
             {
               int x = (c-'0');
-              c = minilisp_getc();
+              c = io->fgetc(io);
               if (c>='0' && c<='7')
                 {
                   x = (x<<3)+(c-'0');
-                  c = minilisp_getc();
+                  c = io->fgetc(io);
                   if (c>='0' && c<='7')
                     {
                       x = (x<<3)+(c-'0');
-                      c = minilisp_getc();
+                      c = io->fgetc(io);
                     }
                 }
               append((char)x, s, l, m);
@@ -1321,22 +1460,22 @@ read_c_string(int &c)
             {
               int x = 0;
               int d = c;
-              c = minilisp_getc();
+              c = io->fgetc(io);
               if (isxdigit(c))
                 {
                   x = (x<<4) + (isdigit(c) ? c-'0' : toupper(c)-'A'+10);
-                  c = minilisp_getc();
+                  c = io->fgetc(io);
                   if (isxdigit(c))
                     {
                       x = (x<<4) + (isdigit(c) ? c-'0' : toupper(c)-'A'+10);
-                      c = minilisp_getc();
+                      c = io->fgetc(io);
                     }
                   append((char)x, s, l, m);
                   continue;
                 }
               else
                 {
-                  minilisp_ungetc(c);
+                  io->ungetc(io, c);
                   c = d;
                 }
             }
@@ -1349,16 +1488,16 @@ read_c_string(int &c)
           }
         }
       append(c,s,l,m);
-      c = minilisp_getc();
+      c = io->fgetc(io);
     }
-  c = minilisp_getc();
+  c = io->fgetc(io);
   r = miniexp_string(s ? s : "");
   delete [] s;
   return r;
 }
 
 static miniexp_t
-read_quoted_symbol(int &c)
+read_quoted_symbol(miniexp_io_t *io, int &c)
 {
   miniexp_t r;
   char *s = 0;
@@ -1367,21 +1506,21 @@ read_quoted_symbol(int &c)
   ASSERT(c == '|');
   for(;;)
     {
-      c = minilisp_getc();
-      if (c==EOF || isascii(c) && !isprint(c))
-        return read_error(c);
+      c = io->fgetc(io);
+      if (c==EOF || (isascii(c) && !isprint(c)))
+        return read_error(io, c);
       if (c=='|')
         break;
       append(c,s,l,m);
     }
-  c = minilisp_getc();
+  c = io->fgetc(io);
   r = miniexp_symbol(s ? s : "");
   delete [] s;
   return r;
 }
 
 static miniexp_t
-read_symbol_or_number(int &c)
+read_symbol_or_number(miniexp_io_t *io, int &c)
 {
   miniexp_t r;
   char *s = 0;
@@ -1389,15 +1528,16 @@ read_symbol_or_number(int &c)
   int m = 0;
   for(;;)
     {
-      if (c==EOF || c=='(' || c==')' || c=='|' || c=='\"'  || 
-          isspace(c) || !isascii(c) || !isprint(c) || 
-          minilisp_macrochar_parser[c] )
+      if (c==EOF || c=='(' || c==')' || c=='|' || c=='\"'  
+          || isspace(c) || !isascii(c) || !isprint(c) 
+          || (io->p_macrochar && io->p_macroqueue  
+              && c < 128 && c >= 0 && io->p_macrochar[c] ) )
         break;
       append(c,s,l,m);
-      c = minilisp_getc();
+      c = io->fgetc(io);
     }
   if (l <= 0)
-    return read_error(c);
+    return read_error(io, c);
   char *end;
   long x = strtol(s, &end, 0);
   if (*end)
@@ -1409,24 +1549,24 @@ read_symbol_or_number(int &c)
 }
 
 static miniexp_t
-read_miniexp(int &c)
+read_miniexp(miniexp_io_t *io, int &c)
 {
   for(;;)
     {
-      if (miniexp_consp(inputqueue))
+      if (io->p_macroqueue && miniexp_consp(*io->p_macroqueue))
         {
-          miniexp_t p = car(inputqueue);
-          inputqueue = cdr(inputqueue);
+          miniexp_t p = car(*io->p_macroqueue);
+          *io->p_macroqueue = cdr(*io->p_macroqueue);
           return p;
         }
-      skip_blank(c);
+      skip_blank(io, c);
       if (c == EOF)
         {
-          return read_error(c);
+          return read_error(io, c);
         }
       else if (c == ')')
         {
-          c = minilisp_getc();
+          c = io->fgetc(io);
           continue;
         }
       else if (c == '(')
@@ -1434,20 +1574,20 @@ read_miniexp(int &c)
           minivar_t l;
           miniexp_t *where = &l;
           minivar_t p;
-          c = minilisp_getc();
+          c = io->fgetc(io);
           for(;;)
             {
-              skip_blank(c);
+              skip_blank(io, c);
               if (c == ')')
                 break;
               if (c == '.')
                 {
-                  int d = minilisp_getc();
-                  minilisp_ungetc(d);
+                  int d = io->fgetc(io);
+                  io->ungetc(io, d);
                   if (isspace(d)) 
                     break;
                 }
-              p = read_miniexp(c);
+              p = read_miniexp(io, c);
               if ((miniexp_t)p == miniexp_dummy)
                 return miniexp_dummy;
               *where = miniexp_cons(p, miniexp_nil);
@@ -1455,50 +1595,98 @@ read_miniexp(int &c)
             }
           if (c == '.')
             {
-              c = minilisp_getc();
-              skip_blank(c);
+              c = io->fgetc(io);
+              skip_blank(io, c);
               if (c != ')')
-                *where = read_miniexp(c);
+                *where = read_miniexp(io, c);
             }
-          skip_blank(c);
+          skip_blank(io, c);
           if (c != ')')
-            return read_error(c);
-          c = minilisp_getc();
+            return read_error(io, c);
+          c = io->fgetc(io);
           return l;
         }
       else if (c == '"')
         {
-          return read_c_string(c);
+          return read_c_string(io, c);
         }
       else if (c == '|')
         {
-          return read_quoted_symbol(c);
+          return read_quoted_symbol(io, c);
         }
-      else if (c>=0 && c<128 && minilisp_macrochar_parser[c])
+      else if (c == '#')
         {
-          miniexp_t p = minilisp_macrochar_parser[c]();
+          return read_error(io, c);
+        }
+      else if (io->p_macrochar && io->p_macroqueue 
+               && c >= 0 && c < 128 
+               && io->p_macrochar[c])
+        {
+          miniexp_t p = io->p_macrochar[c](io);
           if (miniexp_length(p) > 0)
-            inputqueue = p;
-          c = minilisp_getc();
+            *io->p_macroqueue = p;
+          c = io->fgetc(io);
           continue;
         }
       else 
         {
-          return read_symbol_or_number(c);
+          return read_symbol_or_number(io, c);
         }
     }
 }
 
 miniexp_t 
-miniexp_read(void)
+miniexp_read_r(miniexp_io_t *io)
 {
-  int c = minilisp_getc();
-  miniexp_t p = read_miniexp(c);
-  minilisp_ungetc(c);
+  int c = io->fgetc(io);
+  miniexp_t p = read_miniexp(io, c);
+  io->ungetc(io, c);
   return p;
 }
 
 
+/* ---- COMPAT */
+
+miniexp_t miniexp_read(void)
+{
+  return miniexp_read_r(&miniexp_io);
+}
+
+miniexp_t miniexp_prin(miniexp_t p)
+{
+  return miniexp_prin_r(&miniexp_io, p);
+}
+
+miniexp_t miniexp_print(miniexp_t p)
+{
+  return miniexp_print_r(&miniexp_io, p);
+}
+
+miniexp_t miniexp_pprin(miniexp_t p, int w)
+{
+  return miniexp_pprin_r(&miniexp_io, p, w);
+}
+
+miniexp_t miniexp_pprint(miniexp_t p, int w)
+{
+  return miniexp_pprint_r(&miniexp_io, p, w);
+}
+
+void 
+minilisp_set_output(FILE *f)
+{
+  miniexp_io_set_output(&miniexp_io, f);
+}
+
+void 
+minilisp_set_input(FILE *f)
+{
+  miniexp_io_set_input(&miniexp_io, f);
+}
+
+
+
+
 /* -------------------------------------------------- */
 /* CLEANUP (SEE GC ABOVE)                             */
 /* -------------------------------------------------- */
